// BufferedRandomAccessFile.java

package com.ureach.detorres.util;

import java.io.*;

/**
 * This class provides an unsynchronized, buffered RandomAccessFile.  It is
 * way, way, way faster (like 25 or 50 times faster) than RandomAccessFile
 * even with a small buffer size of 1024 bytes.  This class is from
 * JavaWorld Java Tip 26.
 *
 * @author Nick Zhang
 */

public class BufferedRandomAccessFile extends RandomAccessFile
{
  private byte[] m_buffer; // buffered data
  private int m_bufEnd = 0; // num valid bytes in buffer
  private int m_bufPos = 0; // "next" byte in buffer
  private long m_realPos = 0L; // position in file of buffer end
  
  /**
   * Creates a new BufferedRandomAccessFile.
   * @param fileName
   * @param mode
   * @param bufSize
   * @throws java.io.IOException
   */
  
  public BufferedRandomAccessFile (final String fileName, final String mode,
    final int bufSize)
    throws IOException
  {
    super (fileName, mode);
    invalidate ();
    m_buffer = new byte[bufSize];
  }
  
  /**
   * Returns the next byte in the file.
   *
   * @return the next byte in the lower 8 bits
   * @throws IOException on any read error
   */
  
  public final int read ()
    throws IOException
  {
    // Check for need to read more data from disk.
    
    if (m_bufPos >= m_bufEnd)
    {
      if (fillBuffer () < 0)
      {
        return -1;
      }
    }
    
    // Check for no data available.  If data's available, return it in the
    // lower 8 bits of an int.
    
    if (m_bufEnd == 0)
    {
      return -1;
    }
    else
    {
      return m_buffer[m_bufPos++] & 0xff;
    }
  }
  
  /**
   * Reads data from the file to try and fill the buffer.
   *
   * @return the number of bytes read into the buffer
   * @throws java.io.IOException
   */
  
  private int fillBuffer ()
    throws IOException
  {
    // Try reading bytes using the superclass.
    
    int numRead = super.read (m_buffer, 0, m_buffer.length);
    
    // If successful, update the file pointer and buffer pointers.
    
    if (numRead >= 0)
    {
      m_realPos += (long) numRead;
      m_bufEnd = numRead;
      m_bufPos = 0;
    }
    
    return numRead;
  }
  
  /**
   * Invalidates the contents of the buffer.  The next attempt to read will
   * require data to be read from the file.
   *
   * @throws java.io.IOException
   */
  
  private void invalidate ()
    throws IOException
  {
    m_bufEnd = 0;
    m_bufPos = 0;
    m_realPos = super.getFilePointer ();
  }
  
  /**
   * Reads a block of bytes into an array.
   *
   * @param b the destination array
   * @param off the offset into buf to start writing bytes
   * @param len the number of bytes to read
   * @return the number of bytes read
   */
  
  public int read (byte[] b, int off, int len)
    throws IOException
  {
    // If there are enough bytes in the buffer, just copy them and update the
    // pointers.
    
    int leftover = m_bufEnd - m_bufPos;
    
    if (len <= leftover)
    {
      System.arraycopy (m_buffer, m_bufPos, b, off, len);
      m_bufPos += len;
      
      return len;
    }
    else
    {
      // Not enough bytes in the buffer, so copy a byte at a time using the
      // single byte read method.  Seems like this could be done with a copy
      // from the buffer plus a call to super.read (b, off, len)...
      
      for (int idx = 0; idx < len; idx++)
      {
        int c = read ();
        
        if (c != -1)
        {
          b[off + idx] = (byte) c;
        }
        else
        {
          return idx;
        }
      }
      
      return len;
    }
  }
  
  /**
   * Returns the virtual file pointer.
   *
   * @return the effective file pointer
   */
  
  public long getFilePointer ()
    throws IOException
  {
    // m_realPos points to the end of the current buffer, so subtract the
    // number of bytes in the buffer and add the buffer position.  If the
    // buffer has been invalidated, m_bufEnd and m_bufPos are both 0, so it
    // returns the real file pointer.
    
    return (m_realPos - (long) (m_bufEnd - m_bufPos));
  }
  
  /**
   * Seeks to a particular byte in the file.
   *
   * @param pos the desired file position
   */
  
  public void seek (long pos)
    throws IOException
  {
    // If the desired position is in the buffer, adjust the buffer pointer.
    // Otherwise, do a real seek and invalidate the buffer.
    
    int n = (int) (m_realPos - pos);
    
    if (n >= 0 && n <= m_bufEnd)
    {
      m_bufPos = m_bufEnd - n;
    }
    else
    {
      super.seek (pos);
      invalidate ();
    }
  }
  
  /**
   * Replaces RandomAccessFile's readLine method which is final.  Reads a line
   * of text (ending in a NEWLINE).
   *
   * @throws java.io.IOException
   * @return the line as a String
   */
  
  public final String getNextLine ()
    throws IOException
  {
    String str;
    
    if (m_bufEnd - m_bufPos <= 0)
    {
      if (fillBuffer () < 0)
      {
        throw new IOException ("error in filling buffer!");
      }
    }
    
    int lineend = -1;
    
    for (int idx = m_bufPos; idx < m_bufEnd; idx++)
    {
      if (m_buffer[idx] == '\n')
      {
        lineend = idx;
        break;
      }
    }
    
    if (lineend < 0)
    {
      StringBuffer input = new StringBuffer (256);
      int c;
      
      while (((c = read ()) != -1) && (c != '\n'))
      {
        input.append ((char) c);
      }
      
      if ((c == -1) && (input.length () == 0))
      {
        return null;
      }
      
      return input.toString ();
    }
    
    if (lineend > 0 && m_buffer[lineend - 1] == '\r')
    {
      str = new String (m_buffer, 0, m_bufPos, lineend - m_bufPos - 1);
    }
    else
    {
      str = new String (m_buffer, 0, m_bufPos, lineend - m_bufPos);
    }
    
    m_bufPos = lineend + 1;
    
    return str;
  }
}


// eof "BufferedRandomAccessFile.java"
