// HexView.java
// Copyright 2004 by Andrew de Torres
// Minor revisions by Stephen Olano

package com.ureach.detorres.util;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;

/**
 * Simple hexadecimal file viewer.
 *
 * @author Andy de Torres, detorres@ureach.com
 */

public class HexView extends Object {
    public static final int DISPLAY_HEX = 1;
    public static final int DISPLAY_DECIMAL = 2;

    public static final int LITTLE_ENDIAN = 1;
    public static final int BIG_ENDIAN = 2;

    private JFrame frame;
    /** file length */
    private long dataLength;
    /** */
    private long cursorAddress;
    private BufferedRandomAccessFile file;
    private HexViewComponent hexViewComponent;
    private JScrollBar vbar;
    private JTextField jTxtFShort;
    private JTextField jTxtFInt;
    private JTextField jTxtFLong;
    private JTextField jTxtFOffset;
    private JTextField jTxtFFloat;
    private JTextField jTxtFDouble;
    private JToggleButton jRBtnLittleEndian;
    private JToggleButton jRBtnBigEndian;
    private JToggleButton jRBtnHexDisplay;
    private JToggleButton jRBtnDecimalDisplay;
    private Timer displayTimer;
    private boolean isBigEndian = false;
    private boolean isDecimalDisplay = true;

    private final static boolean STANDALONE = true;

    // use for standalone
    public HexView(final String fileName)
            throws FileNotFoundException, IOException {
        Initialize(new FileName(fileName, null), STANDALONE);
    }

    // use for plugin
    public HexView(final FileName fileName)
            throws FileNotFoundException, IOException {
        Initialize(fileName, !STANDALONE);
    }

    public void Initialize(final FileName fileName, final boolean isStandAlone)
            throws FileNotFoundException, IOException {
        file = new BufferedRandomAccessFile(fileName.getFullPath(), "r", 1024);
        dataLength = file.length();

        hexViewComponent = new HexViewComponent() {
            public void setTopAddress(final long topAddress) {
                super.setTopAddress(topAddress);
                vbar.setValue((int) (topAddress >> 4));
            }
        };

        hexViewComponent.setBackground(Color.white);
        hexViewComponent.setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
        createFileDropTarget();

        vbar = new JScrollBar(JScrollBar.VERTICAL, 0, 0, 0,
                (int) ((dataLength + 15L) >> 4));

        vbar.addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(final AdjustmentEvent evt) {
                int row = evt.getValue();
                hexViewComponent.setTopAddress((long) (row << 4));
                hexViewComponent.repaint();
            }
        });

        hexViewComponent.addComponentListener(new ComponentAdapter() {
            public void componentResized(final ComponentEvent evt) {
                vbar.setVisibleAmount(hexViewComponent.getVisibleRowCount());
            }
        });

        JPanel readOutPanel = createReadoutPanel();

        // display filename excluding absolute path
        frame = new JFrame("HexView: " + fileName.getName() + " - Use Arrow Keys to Navigate");
        JPanel contentPane = (JPanel) frame.getContentPane();
        contentPane.setLayout(new BorderLayout(0, 5));
        contentPane.add(hexViewComponent, BorderLayout.CENTER);
        contentPane.add(vbar, BorderLayout.EAST);
        contentPane.add(readOutPanel, BorderLayout.SOUTH);

        frame.setSize(600, 400);

        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(final WindowEvent evt) {
                try {
                    file.close();
                } catch (IOException e) {
                    //ok
                }
                frame.dispose();
                if (isStandAlone) System.exit(0);
            }
        });

        setDisplay(DISPLAY_DECIMAL);
        setEndian(LITTLE_ENDIAN);

        displayTimer = new Timer(150, new ActionListener() {
            public void actionPerformed(final ActionEvent evt) {
                updateConversions();
            }
        });

        displayTimer.setRepeats(false);
        displayTimer.start();

        jRBtnBigEndian.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent evt) {
                setEndian(BIG_ENDIAN);
            }
        });

        jRBtnLittleEndian.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent evt) {
                setEndian(LITTLE_ENDIAN);
            }
        });

        jRBtnDecimalDisplay.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent evt) {
                setDisplay(DISPLAY_DECIMAL);
            }
        });

        jRBtnHexDisplay.addActionListener(new ActionListener() {
            public void actionPerformed(final ActionEvent evt) {
                setDisplay(DISPLAY_HEX);
            }
        });
    }

    public JFrame getFrame() {
        return frame;
    }

    private JPanel createReadoutPanel() {
        Font readoutFont = new Font("Courier New", Font.PLAIN, 12);
        JLabel jLblShort = new JLabel("short:", JLabel.RIGHT);
        jTxtFShort = createReadoutField(8, readoutFont);
        JLabel jLblInt = new JLabel("int:", JLabel.RIGHT);
        jTxtFInt = createReadoutField(15, readoutFont);
        JLabel jLblLong = new JLabel("long:", JLabel.RIGHT);
        jTxtFLong = createReadoutField(24, readoutFont);
        JLabel jLblOffset = new JLabel("offset:", JLabel.RIGHT);
        jTxtFOffset = createReadoutField(8, readoutFont);
        JLabel jLblFloat = new JLabel("float:", JLabel.RIGHT);
        jTxtFFloat = createReadoutField(15, readoutFont);
        JLabel jLblDouble = new JLabel("double:", JLabel.RIGHT);
        jTxtFDouble = createReadoutField(24, readoutFont);

        jRBtnLittleEndian = new JRadioButton("LE");
        jRBtnBigEndian = new JRadioButton("BE");
        jRBtnHexDisplay = new JRadioButton("HEX");
        jRBtnDecimalDisplay = new JRadioButton("DEC");

        ButtonGroup endianGroup = new ButtonGroup();
        endianGroup.add(jRBtnLittleEndian);
        endianGroup.add(jRBtnBigEndian);

        ButtonGroup displayGroup = new ButtonGroup();
        displayGroup.add(jRBtnHexDisplay);
        displayGroup.add(jRBtnDecimalDisplay);

        JPanel readOutPanel = new JPanel(new GridLayout(2, 1, 0, 2));

        final FlowLayout layout = new FlowLayout(FlowLayout.LEFT, 5, 0);
        JPanel panel1 = new JPanel(layout);
        panel1.add(jLblShort);
        panel1.add(jTxtFShort);
        panel1.add(jLblInt);
        panel1.add(jTxtFInt);
        panel1.add(jLblLong);
        panel1.add(jTxtFLong);
        panel1.add(jRBtnLittleEndian);
        panel1.add(jRBtnDecimalDisplay);

        JPanel panel2 = new JPanel(layout);
        panel2.add(jLblOffset);
        panel2.add(jTxtFOffset);
        panel2.add(jLblFloat);
        panel2.add(jTxtFFloat);
        panel2.add(jLblDouble);
        panel2.add(jTxtFDouble);
        panel2.add(jRBtnBigEndian);
        panel2.add(jRBtnHexDisplay);

        readOutPanel.add(panel1);
        readOutPanel.add(panel2);

        makeSamePreferredSize(jLblShort, jLblOffset);
        makeSamePreferredSize(jLblInt, jLblFloat);
        makeSamePreferredSize(jLblLong, jLblDouble);
        makeSamePreferredSize(jRBtnLittleEndian, jRBtnBigEndian);
        makeSamePreferredSize(jRBtnDecimalDisplay, jRBtnHexDisplay);


        return readOutPanel;
    }

    private static JTextField createReadoutField(final int columns,
                                          final Font readoutFont) {
        JTextField tf = new JTextField(columns);
        tf.setFont(readoutFont);
        tf.setHorizontalAlignment(JTextField.RIGHT);

        return tf;
    }

    /** make each have preferred size the max of both */
    private static void makeSamePreferredSize(final JComponent c1, final JComponent c2) {
        Dimension pref1 = c1.getPreferredSize();
        Dimension pref2 = c2.getPreferredSize();
        Dimension pref = new Dimension(Math.max(pref1.width, pref2.width),
                Math.max(pref1.height, pref2.height));

        c1.setPreferredSize(pref);
        c2.setPreferredSize(pref);
    }

    public void setDisplay(final int display) {
        if (display == DISPLAY_DECIMAL) {
            isDecimalDisplay = true;
            jRBtnDecimalDisplay.setSelected(true);
        } else if (display == DISPLAY_HEX) {
            isDecimalDisplay = false;
            jRBtnHexDisplay.setSelected(true);
        } else {
            throw new IllegalArgumentException("display = " + display);
        }

        updateConversions();
    }

    public void setEndian(final int endian) {
        if (endian == BIG_ENDIAN) {
            isBigEndian = true;
            jRBtnBigEndian.setSelected(true);
        } else if (endian == LITTLE_ENDIAN) {
            isBigEndian = false;
            jRBtnLittleEndian.setSelected(true);
        } else {
            throw new IllegalArgumentException("endian = " + endian);
        }

        updateConversions();
    }

    /**
     * Updates the fields that show the value at the cursor location as short,
     * int, long, float, and double.
     */

    private void updateConversions() {
        long curPos = cursorAddress;

        try {
            file.seek(curPos);
            short shortValue;
            int intValue;
            long longValue;
            float floatValue;
            double doubleValue;

            if (isDecimalDisplay) {
                jTxtFOffset.setText(Long.toString(curPos));
            } else {
                jTxtFOffset.setText(Long.toHexString(curPos));
            }

            if (curPos <= dataLength - 8L) {
                byte[] bytes = new byte[8];
                file.readFully(bytes);

                if (isBigEndian) {
                    shortValue = toBigEndianShort(bytes[0], bytes[1]);
                    intValue = toBigEndianInt(bytes[0], bytes[1], bytes[2], bytes[3]);
                    longValue = toBigEndianLong(bytes[0], bytes[1], bytes[2],
                            bytes[3], bytes[4], bytes[5], bytes[6], bytes[7]);
                } else {
                    shortValue = toLittleEndianShort(bytes[0], bytes[1]);
                    intValue = toLittleEndianInt(bytes[0], bytes[1], bytes[2], bytes[3]);
                    longValue = toLittleEndianLong(bytes[0], bytes[1], bytes[2],
                            bytes[3], bytes[4], bytes[5], bytes[6], bytes[7]);
                }

                doubleValue = Double.longBitsToDouble(longValue);
                floatValue = Float.intBitsToFloat(intValue);

                displayLongValue(longValue);
                jTxtFDouble.setText("" + doubleValue);
                displayIntValue(intValue);
                jTxtFFloat.setText("" + floatValue);
                displayShortValue(shortValue);
            } else {
                jTxtFLong.setText("");
                jTxtFDouble.setText("");

                if (curPos <= dataLength - 4L) {
                    byte[] bytes = new byte[4];
                    file.readFully(bytes);

                    if (isBigEndian) {
                        shortValue = toBigEndianShort(bytes[0], bytes[1]);
                        intValue = toBigEndianInt(bytes[0], bytes[1], bytes[2], bytes[3]);
                    } else {
                        shortValue = toLittleEndianShort(bytes[0], bytes[1]);
                        intValue = toLittleEndianInt(bytes[0], bytes[1], bytes[2], bytes[3]);
                    }

                    floatValue = Float.intBitsToFloat(intValue);

                    displayIntValue(intValue);
                    jTxtFFloat.setText("" + floatValue);
                    displayShortValue(shortValue);
                } else {
                    jTxtFInt.setText("");
                    jTxtFFloat.setText("");

                    if (curPos <= dataLength - 2L) {
                        byte[] bytes = new byte[2];
                        file.readFully(bytes);

                        if (isBigEndian) {
                            shortValue = toBigEndianShort(bytes[0], bytes[1]);
                        } else {
                            shortValue = toLittleEndianShort(bytes[0], bytes[1]);
                        }

                        displayShortValue(shortValue);
                    } else {
                        jTxtFShort.setText("");
                    }
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void displayShortValue(final short shortValue) {
        if (isDecimalDisplay) {
            jTxtFShort.setText(Short.toString(shortValue));
        } else {
            jTxtFShort.setText("0x" + Integer.toHexString(shortValue & 0xffff));
        }
    }

    private void displayLongValue(final long longValue) {
        if (isDecimalDisplay) {
            jTxtFLong.setText(Long.toString(longValue));
        } else {
            jTxtFLong.setText("0x" + Long.toHexString(longValue));
        }
    }

    private void displayIntValue(final int intValue) {
        if (isDecimalDisplay) {
            jTxtFInt.setText(Integer.toString(intValue));
        } else {
            jTxtFInt.setText("0x" + Integer.toHexString(intValue));
        }
    }

    private DropTarget createFileDropTarget() {
        DropTarget dropTarget = new DropTarget(hexViewComponent,
                DnDConstants.ACTION_COPY_OR_MOVE,
                new DropTargetAdapter() {
                    public void drop(final DropTargetDropEvent evt) {
                        try {
                            DataFlavor stringFlavor = DataFlavor.stringFlavor;
                            Transferable tr = evt.getTransferable();

                            if (evt.isDataFlavorSupported(stringFlavor)) {
                                evt.acceptDrop(DnDConstants.ACTION_COPY);
                                String fileName = (String) tr.getTransferData(stringFlavor);
                                changeFile(fileName);
                                evt.dropComplete(true);
                            } else if (evt.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
                                evt.acceptDrop(DnDConstants.ACTION_COPY);
                                List files = (List) tr.getTransferData(DataFlavor.javaFileListFlavor);

                                if (files.size() == 1) {
                                    String fileName = files.get(0).toString();
                                    changeFile(fileName);
                                    evt.dropComplete(true);
                                } else {
                                    evt.dropComplete(false);
                                }
                            } else {
                                evt.rejectDrop();
                            }
                        } catch (IOException ioe) {
                            ioe.printStackTrace();
                        } catch (UnsupportedFlavorException ufe) {
                            ufe.printStackTrace();
                        }
                    }

                    private void changeFile(final String fileName)
                            throws IOException {
                        try {
                            file.close();
                        } catch (IOException e) {
                            //ok
                        }

                        file = new BufferedRandomAccessFile(fileName, "r", 1024);
                        dataLength = file.length();
                        vbar.setMaximum((int) ((dataLength + 15L) >> 4));
                        frame.setTitle("HexView: " + fileName);
                        hexViewComponent.setTopAddress(0L);
                        cursorAddress = 0L;
                        vbar.setValue(0);
                        hexViewComponent.repaint();
                        updateConversions();
                    }
                });

        return dropTarget;
    }

    public static short toBigEndianShort(byte b1, byte b2) {
        return (short) (((b1 & 0xff) << 8) | (b2 & 0xff));
    }

    public static int toBigEndianInt(byte b1, byte b2, byte b3, byte b4) {
        return (b1 & 0xff) << 24 | (b2 & 0xff) << 16 | (b3 & 0xff) << 8 | (b4 & 0xff);
    }

    public static long toBigEndianLong(byte b1, byte b2, byte b3, byte b4,
                                       byte b5, byte b6, byte b7, byte b8) {
        return (b1 & 0xffL) << 56 | (b2 & 0xffL) << 48 | (b3 & 0xffL) << 40 | (b4 & 0xffL) << 32
                | (b5 & 0xffL) << 24 | (b6 & 0xffL) << 16 | (b7 & 0xffL) << 8 | (b8 & 0xffL);
    }

    public static short toLittleEndianShort(byte b1, byte b2) {
        return (short) (((b2 & 0xff) << 8) | (b1 & 0xff));
    }

    public static int toLittleEndianInt(byte b1, byte b2, byte b3, byte b4) {
        return (b4 & 0xff) << 24 | (b3 & 0xff) << 16 | (b2 & 0xff) << 8 | (b1 & 0xff);
    }

    public static long toLittleEndianLong(byte b1, byte b2, byte b3, byte b4,
                                          byte b5, byte b6, byte b7, byte b8) {
        return (b8 & 0xffL) << 56 | (b7 & 0xffL) << 48 | (b6 & 0xffL) << 40 | (b5 & 0xffL) << 32
                | (b4 & 0xffL) << 24 | (b3 & 0xffL) << 16 | (b2 & 0xffL) << 8 | (b1 & 0xffL);
    }

    public static void main(String[] args) {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            //ignore
        }

        try {
            HexView hexview = null;

            if (args.length == 0) {
                JFileChooser fileChooser = new JFileChooser(File.separator);
                fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
                fileChooser.setDialogTitle("Select file to view");
                int response = fileChooser.showDialog(new JFrame(), "Open");

                if (response != JFileChooser.APPROVE_OPTION) {
                    System.exit(-1);
                }

                String fileName = fileChooser.getSelectedFile().getAbsolutePath();
                hexview = new HexView(fileName);
            } else if (args.length == 1) {
                hexview = new HexView(args[0]);
            } else {
                System.out.println("args: <file name>");
                System.exit(-1);
            }

            hexview.getFrame().setLocationRelativeTo(null);
            hexview.getFrame().show();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, "The file was not found.\n"
                    + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        } catch (IOException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, "The file could not be read.",
                    "Error", JOptionPane.ERROR_MESSAGE);
        }
    }

    class HexViewComponent extends JComponent {
        private final char[] hexChar = {'0', '1', '2', '3', '4', '5', '6',
                                        '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

        private Font font;
        private Color byteColor = Color.getHSBColor(0.00f, 0.1f, 0.9f);
        private Color shortColor = Color.getHSBColor(0.15f, 0.2f, 0.9f);
        private Color intColor = Color.getHSBColor(0.50f, 0.2f, 0.9f);
        private Color longColor = Color.getHSBColor(0.75f, 0.05f, 0.9f);
        private long topAddress;

        public HexViewComponent() {
            setFocusable(true);

            JTextArea textArea = new JTextArea();
            font = new Font("Courier New", textArea.getFont().getStyle(), textArea.getFont().getSize());

            Action downAction = new AbstractAction("down") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress + 16L < dataLength) {
                        cursorAddress += 16L;

                        if (cursorAddress >= topAddress + (long) (getVisibleRowCount() << 4)) {
                            setTopAddress(topAddress + 16L);
                        }

                        HexViewComponent.this.repaint();
                        displayTimer.restart();
                    }
                }
            };

            Action upAction = new AbstractAction("up") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress - 16L >= 0) {
                        cursorAddress -= 16L;

                        if (cursorAddress < topAddress) {
                            setTopAddress(topAddress - 16L);
                        }

                        HexViewComponent.this.repaint();
                        displayTimer.restart();
                    }
                }
            };

            Action rightAction = new AbstractAction("right") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress + 1L < dataLength) {
                        cursorAddress += 1L;

                        if (cursorAddress >= topAddress + (long) (getVisibleRowCount() << 4)) {
                            setTopAddress(topAddress + 16L);
                        }

                        HexViewComponent.this.repaint();
                        displayTimer.restart();
                    }
                }
            };

            Action leftAction = new AbstractAction("left") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress - 1L >= 0) {
                        cursorAddress -= 1L;

                        if (cursorAddress < topAddress) {
                            setTopAddress(topAddress - 16L);
                        }

                        HexViewComponent.this.repaint();
                        displayTimer.restart();
                    }
                }
            };

            Action pageDownAction = new AbstractAction("pagedown") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress + ((long) (getVisibleRowCount() << 4)) < dataLength) {
                        cursorAddress += (long) (getVisibleRowCount() << 4);
                        displayTimer.restart();
                    }

                    long newTop = Math.min(topAddress + (long) (getVisibleRowCount() << 4),
                            (((dataLength + 15L) >> 4) - (long) getVisibleRowCount()) << 4);

                    newTop = Math.max(newTop, 0L);
                    setTopAddress(newTop);

                    HexViewComponent.this.repaint();
                }
            };

            Action pageUpAction = new AbstractAction("pageup") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress - ((long) (getVisibleRowCount() << 4)) >= 0) {
                        cursorAddress -= (long) (getVisibleRowCount() << 4);
                        displayTimer.restart();
                    }

                    setTopAddress(Math.max(topAddress - (long) (getVisibleRowCount() << 4), 0L));

                    HexViewComponent.this.repaint();
                }
            };

            Action superPageDownAction = new AbstractAction("superpagedown") {
                public void actionPerformed(final ActionEvent evt) {
                    long newTop = Math.min(topAddress + 0x10000L,
                            (((dataLength + (long)15) >> 4) - (long) getVisibleRowCount()) << 4);

                    if (cursorAddress + 0x10000L < dataLength) {
                        cursorAddress += 0x10000L;
                        displayTimer.restart();
                    } else {
                        cursorAddress = newTop + (cursorAddress - topAddress);
                    }

                    newTop = Math.max(newTop, 0L);
                    setTopAddress(newTop);

                    HexViewComponent.this.repaint();
                }
            };

            Action superPageUpAction = new AbstractAction("superpageup") {
                public void actionPerformed(final ActionEvent evt) {
                    if (cursorAddress - 0x10000L >= 0) {
                        cursorAddress -= 0x10000L;
                        displayTimer.restart();
                    } else {
                        cursorAddress = cursorAddress - topAddress;
                    }

                    setTopAddress(Math.max(topAddress - 0x10000L, 0L));

                    HexViewComponent.this.repaint();
                }
            };

            Action endAction = new AbstractAction("end") {
                public void actionPerformed(final ActionEvent evt) {
                    displayTimer.restart();
                    long newTop = (((dataLength + (long)15) >> 4)
                                - (long) getVisibleRowCount()) << 4;
                    newTop = Math.max(newTop, 0L);
                    cursorAddress = newTop + (cursorAddress - topAddress);
                    setTopAddress(newTop);

                    HexViewComponent.this.repaint();
                }
            };

            Action homeAction = new AbstractAction("start") {
                public void actionPerformed(final ActionEvent evt) {
                    displayTimer.restart();
                    cursorAddress = 0L;
                    setTopAddress(0L);
                    HexViewComponent.this.repaint();
                }
            };

            ActionMap actionMap = new ActionMap();
            actionMap.setParent(getActionMap());

            InputMap inputMap = new ComponentInputMap(this);
            inputMap.setParent(getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW));

            registerAction(downAction, actionMap, inputMap, KeyEvent.VK_DOWN);
            registerAction(upAction, actionMap, inputMap, KeyEvent.VK_UP);
            registerAction(rightAction, actionMap, inputMap, KeyEvent.VK_RIGHT);
            registerAction(leftAction, actionMap, inputMap, KeyEvent.VK_LEFT);
            registerAction(pageDownAction, actionMap, inputMap, KeyEvent.VK_PAGE_DOWN);
            registerAction(pageUpAction, actionMap, inputMap, KeyEvent.VK_PAGE_UP);
            registerAction(superPageDownAction, actionMap, inputMap, KeyEvent.VK_PAGE_DOWN, KeyEvent.CTRL_DOWN_MASK);
            registerAction(superPageUpAction, actionMap, inputMap, KeyEvent.VK_PAGE_UP, KeyEvent.CTRL_DOWN_MASK);
            registerAction(homeAction, actionMap, inputMap, KeyEvent.VK_HOME);
            registerAction(endAction, actionMap, inputMap, KeyEvent.VK_END);

            setActionMap(actionMap);
            setInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW, inputMap);
//            addMouseListener(new MouseAdapter() {
//                public void mouseClicked(MouseEvent e) {
//                    final int x = e.getX();
//                    final int y = e.getY();
//                    cursorAddress = ...
//                    topAddress = ...
//                }
//            });
        }

        public void setTopAddress(final long topAddress) {
            this.topAddress = topAddress;
        }

        private void registerAction(final Action downAction,
                                    final ActionMap _actionMap, final InputMap inputMap, final int key,
                                    final int modifiers) {
            _actionMap.put(downAction.getValue(Action.NAME), downAction);
            inputMap.put(KeyStroke.getKeyStroke(key, modifiers),
                    downAction.getValue(Action.NAME));
        }

        private void registerAction(final Action downAction,
                                    final ActionMap _actionMap, final InputMap inputMap,
                                    final int key) {
            registerAction(downAction, _actionMap, inputMap, key, 0);
        }

        private Insets insets = new Insets(0, 0, 0, 0);
        private int fontHeight = -1;
        private int fontWidth;
        private int fontAscent;

        private int getFontHeight() {
            if (fontHeight == -1) {
                fontHeight = getGraphics().getFontMetrics(font).getHeight();
                fontWidth = getGraphics().getFontMetrics(font).charWidth('0');
                fontAscent = getGraphics().getFontMetrics(font).getAscent();
            }

            return fontHeight;
        }

        public int getVisibleRowCount() {
            getInsets(insets);
            return (getHeight() - 2 - insets.top - insets.bottom) / getFontHeight();
        }

        public void paint(final Graphics g) {
            Dimension size = getSize();
            int width = size.width;
            int height = size.height;
            getInsets(insets);

            g.setColor(getBackground());
            g.fillRect(0, 0, width, height);

            Border border = getBorder();

            if (border != null) {
                border.paintBorder(this, g, 0, 0, width, height);
            }

            Color foreground = getForeground();
            g.setFont(font);
            int numRows = getVisibleRowCount();

            try {
                file.seek(topAddress);

                for (int row = 0; row < numRows; row++) {
                    long address = topAddress + (long) (row << 4);

                    if (address >= dataLength) {
                        break;
                    }

                    StringBuffer buffer = new StringBuffer();
                    buffer.append(hexChar[nibble(address, 7)]);
                    buffer.append(hexChar[nibble(address, 6)]);
                    buffer.append(hexChar[nibble(address, 5)]);
                    buffer.append(hexChar[nibble(address, 4)]);
                    buffer.append(hexChar[nibble(address, 3)]);
                    buffer.append(hexChar[nibble(address, 2)]);
                    buffer.append(hexChar[nibble(address, 1)]);
                    buffer.append(hexChar[nibble(address, 0)]);
                    buffer.append(":  ");
                    int[] rowBytes = new int[16];

                    for (int idx = 0; idx < 16; idx++) {
                        if (idx == 8) {
                            buffer.append(' ');
                        }
                        if (address + (long)idx < dataLength) {
                            int value = file.read();
                            buffer.append(hexChar[value >> 4]).append(hexChar[value & 0xf]);
                            rowBytes[idx] = value;
                        } else {
                            buffer.append("  ");
                        }

                        buffer.append(' ');
                    }

                    buffer.append("  ");

                    for (int idx = 0; idx < 16; idx++) {
                        if (address + (long)idx < dataLength) {
                            if (idx == 8) {
                                buffer.append(' ');
                            }
                            if (rowBytes[idx] >= 0x20 && rowBytes[idx] <= 0x7f) {
                                buffer.append((char) rowBytes[idx]);
                            } else {
                                buffer.append('.');
                            }
                        }
                    }

                    int cursorOffset = (int) (cursorAddress - address);
                    int y = fontHeight * row + insets.top + 2 - 1;
                    highlightByte(g, cursorOffset, 0, byteColor, y);
                    highlightByte(g, cursorOffset, 1, shortColor, y);
                    highlightByte(g, cursorOffset, 2, intColor, y);
                    highlightByte(g, cursorOffset, 3, intColor, y);
                    highlightByte(g, cursorOffset, 4, longColor, y);
                    highlightByte(g, cursorOffset, 5, longColor, y);
                    highlightByte(g, cursorOffset, 6, longColor, y);
                    highlightByte(g, cursorOffset, 7, longColor, y);

                    g.setColor(foreground);
                    g.drawString(buffer.toString(), insets.left + 2,
                            fontAscent + fontHeight * row + insets.top + 2);
                }
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }
        }

        private void highlightByte(final Graphics g,
                                   final int cursorOffset, int idx, final Color color, final int _y) {
            if ((cursorAddress + (long)idx < dataLength)
                    && (cursorAddress + (long)idx < topAddress + (long)(getVisibleRowCount() << 4))) {
                final int effctvOff = cursorOffset + idx;
                int gapCompensation = 0;
                if (effctvOff >= 0 && effctvOff < 16) {
                    g.setColor(color);
                    if (effctvOff > 7) {
                        gapCompensation = 1;
                    }
                    g.fillRect((effctvOff * 3 + 11 + gapCompensation) * fontWidth + insets.left + 2 - 1, _y, (fontWidth << 1) + 2, fontHeight);
                    g.fillRect((effctvOff + 62 + gapCompensation) * fontWidth + insets.left + 2, _y, fontWidth, fontHeight);
                }
            }
        }

        private byte nibble(final long _x, final int idx) {
            return (byte) ((_x >> (idx << 2)) & 0xf);
        }
    }
}


// eof "HexView.java"
